/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.behavior;

import java.util.Set;


import jp.digitalmuseum.mr.core.BehaviorAbstractImpl;
import jp.digitalmuseum.mr.core.Resource;
import jp.digitalmuseum.mr.resource.Wheels;
import jp.digitalmuseum.mr.resource.Wheels.SPIN;
import jp.digitalmuseum.utils.MathUtils;
import jp.digitalmuseum.utils.Position;
import jp.digitalmuseum.utils.Vector2D;


/**
 * Abstract behavior that only needs getVector(Position) method
 * implemented by sub-classes to work.
 *
 * @author Jun KATO
 */
public abstract class VectorBehavior extends LocationBasedBehaviorAbstractImpl {
	final private Vector2D vector = new Vector2D();
	private Wheels wheels;
	private double allowedDeviationAngle = 0.3; // 17degrees+
	private double allowedNorm = 5.0;
	private double previousNorm;

	/**
	 * @return Returns a set of required interfaces (including DifferentialWheels) for this behavior.
	 */
	@Override
	public Set<Class<? extends Resource>> getRequirements() {
		Set<Class<? extends Resource>> requirements = super.getRequirements();
		requirements.add(Wheels.class);
		return requirements;
	}

	/**
	 * Implementation class of VectorBehavior need not return DifferentialWheels interface,
	 * because VectorBehavior automatically requires it in getRequirements().
	 * @see #getRequirements()
	 * @see BehaviorAbstractImpl#requirements()
	 */
	protected Class<?>[] requirements() {
		return null;
	}

	@Override
	protected void onStart() {
		super.onStart();
		wheels = (Wheels) getResourceMap().get(Wheels.class);
		previousNorm = Double.MAX_VALUE;
	}

	protected void behave() {

		// Get the current position.
		final Position position = getPosition();
		if (position != null) {

			// Get the vector.
			getVectorOut(position, vector);

			// Check if the vector is short enough.
			final double norm = vector.getNorm();
			if (norm < allowedNorm &&
					norm > previousNorm) {
				if (!transitToDefaultNextBehavior()) {
					wheels.stopWheels();
				}
				return;
			}

			// If the vector is not so short, go on.
			final double
			curDirection = getRotation(),
			destDirection = Math.atan2(vector.getY(), vector.getX()),
			deviationAngle = MathUtils.roundRadian(destDirection - curDirection),
			absDeviationAngle = Math.abs(deviationAngle);

			// Rotate if needed,
			if (absDeviationAngle > allowedDeviationAngle) {
				wheels.spin(deviationAngle > 0 ? SPIN.LEFT : SPIN.RIGHT);

			// otherwise simply go forward.
			} else {
				wheels.goForward();
			}
			previousNorm = norm;
			return;
		}
		wheels.stopWheels();
	}

	/**
	 * @param position
	 * @return Returns a vector that specifies in which direction the assigned robot should go.
	 */
	public Vector2D getVector(Position position) {
		final Vector2D vector = new Vector2D();
		getVectorOut(position, vector);
		return vector;
	}

	public abstract void getVectorOut(Position position, Vector2D vector);

	public void setAllowedDeviationAngle(double allowedDeviationAngle) {
		this.allowedDeviationAngle = allowedDeviationAngle;
	}

	public void setAllowedNorm(double allowedNorm) {
		this.allowedNorm = allowedNorm;
	}

}
