/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.util.HashMap;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import jp.digitalmuseum.mr.core.message.MrBehaviorEvent;
import jp.digitalmuseum.mr.core.message.MrEvent;
import jp.digitalmuseum.mr.core.message.MrEventListener;
import jp.digitalmuseum.mr.core.message.MrRobotEvent;

/**
 * Abstract implementation of Behavior interface.<br />
 * All Behavior implementation classes must extend this abstract class.
 *
 * @author Jun KATO
 */
public abstract class BehaviorAbstractImpl extends ServiceAbstractImpl implements Behavior {
	private Robot robot;
	private Set<Class<? extends Resource>> requirements;
	private HashSet<Transition> transitions;
	private Queue<MrRobotEvent> robotEventQueue;
	private HashMap<Class<? extends Resource>, Resource> resourceMap;
	private BehaviorAbstractImpl child;
	private BehaviorAbstractImpl parent;
	private BehaviorAbstractImpl previous;
	private Behavior next;
	private RobotEventListener robotEventListener;

	@SuppressWarnings("unchecked")
	public BehaviorAbstractImpl() {
		super();
		requirements = new HashSet<Class<? extends Resource>>();
		transitions = new HashSet<Transition>();
		robotEventQueue = new LinkedList<MrRobotEvent>();
		final Class<?>[] requirementsArray = requirements();
		if (requirementsArray != null) {
			for (Class<?> c : requirementsArray) {
				requirements.add((Class<? extends Resource>) c);
			}
		}
		robotEventListener = new RobotEventListener();
	}

	/**
	 * @return Returns a set of required interfaces for this behavior.
	 */
	public Set<Class<? extends Resource>> getRequirements() {
		return new HashSet<Class<? extends Resource>>(requirements);
	}

	public boolean hasAssignedRobot() {
		return robot != null;
	}

	public Robot getAssignedRobot() {
		return robot;
	}

	/**
	 * @return Returns the current resource map.
	 */
	protected HashMap<Class<? extends Resource>, Resource> getResourceMap() {
		checkIfRobotIsAssigned();
		return new HashMap<Class<? extends Resource>, Resource>(resourceMap);
	}

	@Override
	public boolean isStarted() {
		return isDelegate() ?
				parent.isStarted() : super.isStarted();
	}

	@Override
	public boolean isPaused() {
		return isDelegate() ?
				parent.isPaused() : super.isPaused();
	}

	/**
	 * @return Returns if this behavior is delegated a robot by another behavior.
	 */
	private boolean isDelegate() {
		return parent != null;
	}

	/**
	 * @return Returns if this behavior delegates a robot to another behavior.
	 */
	private boolean hasDelegate() {
		return child != null;
	}

	@Override
	final public void start() {
		checkIfRobotIsAssigned();
		super.start();
		onStart();
	}

	@Override
	final public void start(ServiceGroup serviceGroup) {
		checkIfRobotIsAssigned();
		super.start(serviceGroup);
		onStart();
	}

	@Override
	final public void pause() {
		checkIfRobotIsAssigned();
		if (hasDelegate()) {
			child.pause();
		}
		super.pause();
	}

	@Override
	final public void resume() {
		checkIfRobotIsAssigned();
		if (hasDelegate()) {
			child.resume();
		}
		super.resume();
	}

	/**
	 * Free resources and stop this behavior.
	 */
	@Override
	final public void stop() {
		checkIfRobotIsAssigned();

		// Leaf behavior has resources.
		if (hasDelegate()) {
			child.stop();
		} else {
			for (Resource resource : resourceMap.values()) {
				resource.free();
			}
		}
		resourceMap.clear();

		if (isDelegate()) {
			distributeEvent(new MrBehaviorEvent(
					this,
					MrBehaviorEvent.STATUS.FINISHED_DELEGATION));
		}

		receiveRobotEvent(false);
		robot = null;

		// Unregister from the service group.
		super.stop();
	}

	/**
	 * @see ServiceAbstractImpl#dispose()
	 */
	@Override
	final public void dispose() {
		super.dispose();
		if (isStarted()) {
			stop();
		}
	}

	final public void run() {
		synchronized (transitions) {
			for (Transition transition : transitions) {
				if (transition.check(this)) {
					final TransitionPreprocess preprocess = transition.getPreprocess();
					if (preprocess != null) {
						preprocess.run(this, transition);
					}
					transitToBehavior(transition.getNextBehavior());
				}
			}
		}
		if (hasDelegate()) {
			child.behave();
		}
		else {
			behave();
		}
	}

	final public boolean assignToRobot(Robot robot) {
		if (robot == null) {
			return false;
		}

		// Check if the robot provides sufficient resource types.
		final Set<Class<? extends Resource>> supply = robot.getAvailableResourceTypes();
		final Set<Class<? extends Resource>> demand = getRequirements();
		if (!CoreUtils.requirementsSuppliable(demand, supply)) {
			return false;
		}

		// Acquire resources.
		resourceMap = robot.requestResources(this, demand);

		// Setup this behavior.
		this.robot = robot;
		return true;
	}

	final public void setDefaultTransition(Behavior behavior) {
		next = behavior;
	}

	final public void addTransition(Transition transition) {
		synchronized (transitions) {
			transitions.add(transition);
		}
	}

	final public void removeTransition(Transition transition) {
		synchronized (transitions) {
			transitions.remove(transition);
		}
	}

	/**
	 * This method is called only once at the construction of this object.
	 * @return Returns an array of required interfaces for this behavior.
	 */
	protected abstract Class<?>[] requirements();

	protected abstract void behave();

	/**
	 * Called when this behavior starts.
	 * This method is usually used for getting a resource object and saving it to a instance field.
	 */
	protected void onStart() {
		// Do nothing unless overrode by subclasses.
	}

	/**
	 * Called when this behavior resumes from delegation.
	 * @see #delegateToBehavior(Behavior)
	 */
	protected void onResumeFromDelegation(Behavior behavior) {
		// Do nothing unless overrode by subclasses.
	}

	final protected boolean transitToBehavior(Behavior behavior) {
		if (behavior == null) {
			return false;
		}
		checkIfRobotIsAssigned();

		synchronized (robot) {

			// Finish delegation.
			// @see #delegateToBehavior(Behavior)
			if (isDelegate()) {
				if (!behavior.equals(parent)) {
					return false;
				}
				final BehaviorAbstractImpl p = parent;
				parent = null;
				robot = null;
				resourceMap.clear();
				p.child = null;
				p.onResumeFromDelegation(this);
				return true;
			}

			// Start next behavior.
			else {
				if (!startNextBehavior(behavior)) {
					return false;
				}
				stop();
				return true;
			}
		}
	}

	final protected boolean transitToDefaultNextBehavior() {
		return transitToBehavior(next);
	}

	/**
	 * Delegates the assigned robot to the other specified behavior.<br />
	 * When this method succeeds, run() of this class won't be called till the delegated behavior stops.
	 *
	 * @param behavior
	 * @return Returns true if delegation is succeeded.
	 * @see #onResumeFromDelegation(Behavior)
	 */
	final protected boolean delegateToBehavior(Behavior behavior) {
		checkIfRobotIsAssigned();

		synchronized (robot) {
			if (!(behavior instanceof BehaviorAbstractImpl)) {
				throw new IllegalArgumentException("Given behavior ("+behavior.getName()
						+") is not assumed to be used as a delegate.");
			}

			final Set<Class<? extends Resource>> supply = resourceMap.keySet();
			final Set<Class<? extends Resource>> demand = behavior.getRequirements();
			supply.addAll(CoreUtils.getClassesOfResources(resourceMap.values()));
			if (!CoreUtils.requirementsSuppliable(demand, supply)) {
				return false;
			}

			// Setup delegation.
			// @see #transitToBehavior(Behavior)
			child = (BehaviorAbstractImpl) behavior;
			child.parent = this;
			child.robot = this.robot;
			child.resourceMap = CoreUtils.constructResourceMap(demand, resourceMap.values());
			child.onStart();

			// behave() of this behavior will not be called till the delegation ends.
			return true;
		}
	}

	final protected boolean rollbackToBehavior() {
		checkIfRobotIsAssigned();

		synchronized (robot) {
			if (!transitToBehavior(previous)) {
				return false;
			}
			previous = null;
			return true;
		}
	}

	private boolean startNextBehavior(Behavior behavior) {
		if (behavior.hasAssignedRobot()) {
			throw new IllegalStateException("Given behavior ("+behavior.getName()
					+") is already assigned to a robot: "+behavior.getAssignedRobot());
		}

		// Check if the robot provides sufficient resource types.
		final Set<Class<? extends Resource>> supply = robot.getAvailableResourceTypes();
		supply.addAll(CoreUtils.getClassesOfResources(resourceMap.values()));
		final Set<Class<? extends Resource>> demand = behavior.getRequirements();
		if (!CoreUtils.requirementsSuppliable(demand, supply)) {
			return false;
		}

		// Free resources of the robot acquired by this behavior.
		for (Resource resource : resourceMap.values()) {
			resource.free();
		}
		resourceMap.clear();

		// Fire transition to the behavior.
		if (behavior instanceof BehaviorAbstractImpl) {
			((BehaviorAbstractImpl) behavior).previous = this;
		}
		if (behavior.assignToRobot(robot)) {
			behavior.start(getServiceGroup());
			return true;
		}
		return false;
	}

	/**
	 * Listen to/Ignore RobotEvent brought by the assigned robot.
	 * @see #hasNextRobotEvent()
	 * @see #pollRobotEvent()
	 */
	final protected void receiveRobotEvent(boolean flag) {
		if (flag) {
			robot.addEventListener(robotEventListener);
		} else {
			robot.removeEventListener(robotEventListener);
		}
	}

	/**
	 * @return Returns true if at least one RobotEvent is available in the event queue.
	 */
	final protected boolean hasNextRobotEvent() {
		return !robotEventQueue.isEmpty();
	}

	/**
	 * @return Returns RobotEvent if available in the event queue. Otherwise, returns null.
	 */
	final protected MrRobotEvent pollRobotEvent() {
		return robotEventQueue.poll();
	}

	/**
	 * Check if this behavior is assigned to a robot.
	 * @exception IllegalStateException
	 */
	private void checkIfRobotIsAssigned() {
		if (robot == null) {
			throw new IllegalStateException("This behavior ("+getName()+") is not assigned to a robot.");
		}
	}

	private class RobotEventListener implements MrEventListener {

		/**
		 * <b>Note:</b> This method here is thread-safe by natural, because only one robot calls this method.
		 */
		public void eventOccurred(MrEvent event) {
			if (event instanceof MrRobotEvent) {
				robotEventQueue.add((MrRobotEvent) event);
			}
		}
	}
}
