/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.entity;

import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jbox2d.collision.shapes.CircleDef;
import org.jbox2d.collision.shapes.ShapeDef;
import org.jbox2d.common.MathUtils;
import org.jbox2d.common.Vec2;
import org.jbox2d.dynamics.Body;
import org.jbox2d.dynamics.BodyDef;
import org.jbox2d.dynamics.World;

import jp.digitalmuseum.mr.core.Behavior;
import jp.digitalmuseum.mr.core.CoreUtils;
import jp.digitalmuseum.mr.core.HakoniwaEntityAbstractImpl;
import jp.digitalmuseum.mr.core.Resource;
import jp.digitalmuseum.mr.core.Robot;
import jp.digitalmuseum.mr.core.message.MrEvent;
import jp.digitalmuseum.mr.core.message.MrEventListener;
import jp.digitalmuseum.mr.resource.Wheels;
import jp.digitalmuseum.utils.Location;

/**
 * HakoniwaRobot, a robot on the Hakoniwa simulator.
 *
 * @author Jun KATO
 * @see HakoniwaRobotWheels
 */
public class HakoniwaRobot extends HakoniwaEntityAbstractImpl implements Robot {

	/** Radius of the robot in [m] */
	final private static float RADIUS = 0.2f;
	final private static float FRICTION = 0.3f;
	final private static float RESTITUTION = 0.1f;
	final private static float WEIGHT = 0.8f;

	public enum STATUS {
		GO_BACKWARD, GO_FORWARD, SPIN_LEFT, SPIN_RIGHT, STOP
	}

	protected HakoniwaRobotWheels wheels;
	protected Behavior wheelsBehavior;

	/** List of MrEventListener */
	private List<MrEventListener> listeners;

	private Shape shape;
	private Body body;
	private ShapeDef sd;
	protected STATUS status;
	private Color color;

	protected HakoniwaRobot() {
		// Do nothing. This constructor is to be called from a subclass.
	}

	public HakoniwaRobot(String name, Location location) {
		setName(name);
		setup(location);
		registerToHakoniwa();
	}

	protected void setup(Location location) {
		shape = new Ellipse2D.Float(
				-RADIUS*100, -RADIUS*100,
				RADIUS*100, RADIUS*100);

		sd = new CircleDef();
		((CircleDef) sd).radius = RADIUS;
		sd.restitution = RESTITUTION;
		sd.density = WEIGHT / (RADIUS * RADIUS * MathUtils.PI);
		sd.friction = FRICTION;

		final BodyDef bd = new BodyDef();
		bd.position = new Vec2();
		bd.position.x = (float) location.getX()/100;
		bd.position.y = (float) location.getY()/100;
		bd.angle = (float) location.getRotation();

		final World world = getHakoniwa().getWorld();
		synchronized (world) {
			body = world.createBody(bd);
			body.createShape(sd);
			body.setMassFromShapes();
		}
		body.m_userData = this;

		wheels = new HakoniwaRobotWheels(this);
		listeners = new ArrayList<MrEventListener>();
	}

	public void preStep() {
		if (status.equals(STATUS.GO_FORWARD) ||
				status.equals(STATUS.GO_BACKWARD)) {
			final Vec2 force = new Vec2();
			final float angle = body.getAngle();
			final float ex = MathUtils.cos(angle);
			final float ey = MathUtils.sin(angle);

			// Acceleration in [m/s^2] and force in [N].
			/*
			final float a = 1f;
			final float f = body.getMass() * (status.equals(STATUS.GO_FORWARD) ? a : -a);
			*/
			final float f = status.equals(STATUS.GO_FORWARD) ? 0.5f : -0.5f;

			// Apply motor power.
			force.x += f * ex;
			force.y += f * ey;
			body.applyForce(force, body.getPosition());

			// Apply frictional force.
			// This can also be simulated by setting linear damping like body.setLinearDamping(0.01f);
			/*
			final Vec2 v = body.getLinearVelocity();
			final float velocity = v.length();
			final float k = 0.5f;
			final float N = k * body.getMass() * 9.8f;
			final float tmp = N / velocity;
			force.x -= tmp * v.x;
			force.y -= tmp * v.y;
			*/
		}
		else if (status.equals(STATUS.SPIN_LEFT) ||
				status.equals(STATUS.SPIN_RIGHT)) {

			// Apply motor power.
			final float torque = status.equals(STATUS.SPIN_LEFT) ?
					0.03f : -0.03f; // [Nm]
			body.applyTorque(torque);
		}
	}

	public float getRadius() {
		return RADIUS*100;
	}

	public Shape getShape() {
		return shape;
	}

	public Color getColor() {
		return color == null ? Color.orange : color;
	}

	public void setColor(Color color) {
		this.color = color;
	}

	public Body getBody() {
		return body;
	}

	public STATUS getStatus() {
		return status;
	}

	public String getTypeName() {
		return "Hakoniwa Robot";
	}

	public Set<Class<? extends Resource>> getAvailableResourceTypes() {
		final Set<Class<? extends Resource>> resourceTypes = new HashSet<Class<? extends Resource>>();
		if (wheelsBehavior != null) {
			return resourceTypes;
		}
		resourceTypes.add(HakoniwaRobotWheels.class);
		return resourceTypes;
	}

	public Set<Class<? extends Resource>> getResourceTypes() {
		final Set<Class<? extends Resource>> resourceTypes = new HashSet<Class<? extends Resource>>();
		resourceTypes.add(HakoniwaRobotWheels.class);
		return resourceTypes;
	}

	public Behavior getAssignedBehavior(Class<? extends Resource> resourceType) {
		if (CoreUtils.classCastableTo(HakoniwaRobotWheels.class, resourceType)) {
			return wheelsBehavior;
		}
		return null;
	}

	public Map<Class<? extends Resource>, Behavior> getBehaviorMap() {
		HashMap<Class<? extends Resource>, Behavior> behaviorMap =
				new HashMap<Class<? extends Resource>, Behavior>();
		if (wheelsBehavior != null) {
			behaviorMap.put(HakoniwaRobotWheels.class, wheelsBehavior);
		}
		return behaviorMap;
	}

	public HashMap<Class<? extends Resource>, Resource> requestResources(
			Behavior behavior, Set<Class<? extends Resource>> resourceTypes) {
		if (wheelsBehavior != null) {
			return null;
		}

		final Set<Resource> resources = new HashSet<Resource>();
		resources.add(wheels);
		final HashMap<Class<? extends Resource>, Resource> resourceMap =
			CoreUtils.constructResourceMap(resourceTypes, resources);

		// Set "in use" flags if succeeded.
		if (resourceMap != null) {
			wheelsBehavior = behavior;
		}

		return resourceMap;
	}

	/**
	 * Bring the given event to the MrEvent listeners.<br />
	 * <b>Caution:</b> This method should only be called by HakoniwaRobot or its subclasses.
	 *
	 * @param event MrEvent to be brought to listeners.
	 * @see #addEventListener(MrEventListener)
	 * @see #removeEventListener(MrEventListener)
	 */
	protected void distributeEvent(MrEvent event) {
		synchronized (listeners) {
			for (MrEventListener listener : listeners) {
				listener.eventOccurred(event);
			}
		}
	}

	/**
	 * Add a MrEvent listener.
	 *
	 * @param mrEventListener
	 * @see #removeEventListener(MrEventListener)
	 */
	public void addEventListener(MrEventListener mrEventListener) {
		synchronized (listeners) {
			listeners.add(mrEventListener);
		}
	}

	/**
	 * Remove a MrEvent listener.
	 *
	 * @param mrEventListener
	 * @return Returns true if succeeded.
	 * @see #addEventListener(MrEventListener)
	 */
	public boolean removeEventListener(MrEventListener mrEventListener) {
		synchronized (listeners) {
			return listeners.remove(mrEventListener);
		}
	}

	/**
	 * Wheels of HakoniwaRobot.
	 *
	 * @author Jun KATO
	 * @see HakoniwaRobot
	 */
	protected class HakoniwaRobotWheels implements Wheels {

		public HakoniwaRobotWheels(HakoniwaRobot robot) {
			status = STATUS.STOP;
		}

		protected void dispose() {
			stopWheels();
		}

		public void goBackward() {
			status = STATUS.GO_BACKWARD;
		}

		public void goForward() {
			status = STATUS.GO_FORWARD;
		}

		public void spin(SPIN direction) {
			if (direction.equals(SPIN.LEFT)) {
				status = STATUS.SPIN_LEFT;
			} else {
				status = STATUS.SPIN_RIGHT;
			}
		}

		public void stopWheels() {
			status = STATUS.STOP;
		}

		public void free() {
			dispose();
			wheelsBehavior = null;
		}
	}

}
