/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.entity;

import java.awt.Shape;

import java.io.IOException;
import java.io.InputStream;

import jp.digitalmuseum.connector.Connector;
import jp.digitalmuseum.mr.core.Resource;
import jp.digitalmuseum.mr.core.ResourceAbstractImpl;
import jp.digitalmuseum.mr.core.RobotAbstractImpl;
import jp.digitalmuseum.mr.resource.DifferentialWheels;

/**
 * LEGO Mindstorms NXT<br />
 *
 * @author Jun KATO
 * @see MindstormsNXTDifferentialWheels
 */
public class MindstormsNXT extends RobotAbstractImpl {
	Resource[] resources;
	DifferentialWheels dw;

	public MindstormsNXT(String name) {
		super();
		initialize(name);
	}

	public MindstormsNXT(String name, String connectionString) {
		super(connectionString);
		initialize(name);
	}

	public MindstormsNXT(String name, Connector connector) {
		super(connector);
		initialize(name);
	}

	@Override
	public void dispose() {
		dw.stopWheels();
		super.dispose();
	}

	private void initialize(String name) {
		setName(name);
		setTypeName("MindstormNXT");
		dw = new MindstormsNXTDifferentialWheels(this);
		resources = new Resource[] { dw };
	}

	@Override
	protected Resource[] getResources() {
		return resources.clone();
	}

	public Shape getShape() {
		return null;
	}

	/**
	 * Differential wheels of LEGO Mindstorms NXT.<br />
	 * TODO Clean up this class.
	 *
	 * @author Jun KATO
	 * @see MindstormsNXT
	 */
	public static class MindstormsNXTDifferentialWheels extends ResourceAbstractImpl implements DifferentialWheels {
		private  final static byte DIRECT_COMMAND_NOREPLY = (byte)0x80;
		private enum STATUS {
			GO_BACK, GO_FORWARD, SPIN_LEFT, SPIN_RIGHT, STOP
		}
		private STATUS status;

		private final static int MAXIMUM_SPEED = 100;
		private int speed = 30;

		public MindstormsNXTDifferentialWheels(MindstormsNXT mindstormsNXT) {
			super(mindstormsNXT);
			initialize();
		}

		public MindstormsNXTDifferentialWheels(Connector connector) {
			super(connector);
			initialize();
		}

		private void initialize() {
			sendAck();
			status = STATUS.STOP;
		}

		public void goBackward() {
			if (!status.equals(STATUS.GO_BACK)) {
				setOutputState(0xff, (byte)speed, 0x01+0x04, 0x01, 0, 0x20, 0);
				status = STATUS.GO_BACK;
			}
		}

		public void goForward() {
			if (!status.equals(STATUS.GO_FORWARD)) {
				setOutputState(0xff, (byte)-speed, 0x01+0x04, 0x01, 0, 0x20, 0);
				status = STATUS.GO_FORWARD;
			}
		}

		public double getSpeed() {
			return speed;
		}

		/**
		 * speed: 0 - 100
		 */
		public void setSpeed(double speed) {
			int s = (int) Math.abs(speed);
			if(MAXIMUM_SPEED < s) {
				System.err.println("Too fast. Maximum speed is 100.");
			}
			this.speed = s;
		}

		public void spin(SPIN spin) {
			if (spin == SPIN.LEFT) {
				if (status != STATUS.SPIN_LEFT) {
					int s = (int) (speed * 0.9);
					setOutputState(2, (byte)-s, 0x01+0x04, 0x01, 0, 0x20, 0);
					setOutputState(1, (byte) s, 0x01+0x04, 0x01, 0, 0x20, 0);
					status = STATUS.SPIN_LEFT;
				}
			} else {
				if (status != STATUS.SPIN_RIGHT) {
					int s = (int) (speed * 0.9);
					setOutputState(1, (byte)-s, 0x01+0x04, 0x01, 0, 0x20, 0);
					setOutputState(2, (byte) s, 0x01+0x04, 0x01, 0, 0x20, 0);
					status = STATUS.SPIN_RIGHT;
				}
			}
		}

		public void rotate(double radius, SPIN direction) {
			throw new UnsupportedOperationException();
		}

		public void stopWheels() {
			setOutputState(0xff, (byte)0, 0x01+0x02, 0x01, 0, 0x20, 0);
			status = STATUS.STOP;
		}

		public boolean sendAck() {
			byte[] ack = {0x01, (byte) 0x88};
			if(!write(ack))
				return false;

			byte[] ret = read();
			if(0 < ret[2]) {
				return false;
			}
			System.out.println("NXT Firmware version: " + ret[4] + "." + ret[3] + ", " + ret[6] + "." + ret[5]);
			return true;
		}

		/**
		 * @param port - Output port (0 - 2 or 0xFF for all three)
		 * @param power - Setpoint for power. (-100 to 100)
		 * @param mode - Setting the modes MOTORON, BRAKE, and/or REGULATED. This parameter is a bitfield, so to put it in brake mode and regulated, use BRAKEMODE + REGULATED
		 * @param regulationMode - see NXTProtocol for enumerations
		 * @param turnRatio - Need two motors? (-100 to 100)
		 * @param runState - see NXTProtocol for enumerations
		 * @param tachoLimit - Number of degrees(?) to rotate before stopping.
		 */
		private boolean setOutputState(int port, byte power, int mode, int regulationMode, int turnRatio, int runState, int tachoLimit) {
			// !! Needs to check port, power to verify they are correct ranges.
			byte [] request = {DIRECT_COMMAND_NOREPLY, 0x04, (byte)port, power, (byte)mode, (byte)regulationMode, (byte)turnRatio, (byte)runState, (byte)tachoLimit, (byte)(tachoLimit>>>8), (byte)(tachoLimit>>>16), (byte)(tachoLimit>>>24)};
			return write(request);
		}

		private boolean write(byte[] buf) {
			// length of packet (Least and Most significant byte)
			// NOTE Bluetooth only. If do USB, doesn't need it.
			final byte LSB = (byte) buf.length;
			final byte MSB = (byte) (buf.length >>> 8);

			// Send length of packet:
			if (!getConnector().write(LSB)) { return false; }
			if (!getConnector().write(MSB)) { return false; }
			if (!getConnector().write(buf)) { return false; }
			try { Thread.sleep(10); }
			catch (InterruptedException e) {  }
			return true;
		}

		private byte[] read() {
			final InputStream inStream = getConnector().getInputStream();
			byte[] reply = null;
			int length = -1;

			try {
				do {
					length = inStream.read();		// First byte specifies length of packet.
				} while (length < 0);
				int lengthMSB = inStream.read();	// Most Significant Byte value
				length = (0xFF & length) | ((0xFF & lengthMSB) << 8);
				reply = new byte[length];
				int read = inStream.read(reply);
			} catch (IOException e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
			}
			return reply;
		}

		@Override
		protected void dispose() {
			stopWheels();
		}
	}
}
