package jp.digitalmuseum.mr.service;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.util.HashSet;
import java.util.Set;

import jp.digitalmuseum.mr.core.Entity;
import jp.digitalmuseum.mr.core.HakoniwaEntity;
import jp.digitalmuseum.mr.core.HakoniwaEntityHolder;
import jp.digitalmuseum.mr.core.Matereal;
import jp.digitalmuseum.mr.core.ServiceAbstractImpl;
import jp.digitalmuseum.utils.Location;
import jp.digitalmuseum.utils.Position;
import jp.digitalmuseum.utils.ScreenLocation;
import jp.digitalmuseum.utils.ScreenPosition;

import org.jbox2d.collision.AABB;
import org.jbox2d.collision.shapes.CircleShape;
import org.jbox2d.collision.shapes.EdgeShape;
import org.jbox2d.collision.shapes.PointShape;
import org.jbox2d.collision.shapes.PolygonShape;
import org.jbox2d.collision.shapes.Shape;
import org.jbox2d.common.MathUtils;
import org.jbox2d.common.OBBViewportTransform;
import org.jbox2d.common.Settings;
import org.jbox2d.common.Vec2;
import org.jbox2d.common.XForm;
import org.jbox2d.dynamics.Body;
import org.jbox2d.dynamics.World;
import org.jbox2d.dynamics.joints.ConstantVolumeJoint;
import org.jbox2d.dynamics.joints.Joint;
import org.jbox2d.dynamics.joints.MouseJoint;
import org.jbox2d.dynamics.joints.MouseJointDef;
import org.jbox2d.dynamics.joints.PulleyJoint;

public class Hakoniwa extends ServiceAbstractImpl implements LocationProvider, ScreenLocationProvider, CoordProvider {

	/** Default maximum world width. */
	private final static float WORLD_WIDTH = 400f;
	/** Default maximum world height. */
	private final static float WORLD_HEIGHT = 320f;

	/** Default viewport width. [px] */
	private final static int SCREEN_WIDTH = 640;
	/** Default viewport height. [px] */
	private final static int SCREEN_HEIGHT = 480;

	/** Default scaling. 640[px] = 10[m] */
	private final static float SCALE = 64;

	private final static int CIRCLE_SEGMENTS = 16;
	private final static float CIRCLE_THETA_INCREMENT = 2.0f * MathUtils.PI / CIRCLE_SEGMENTS;
	private final static int POINT_SEGMENTS = 5;
	private final static int POINT_RADIUS = 3;
	private final static float POINT_THETA_INCREMENT = 2.0f * MathUtils.PI / POINT_SEGMENTS;

	private final static Color sleepingColor = new Color(0.5f, 0.5f, 0.5f);
	// private final static Color selectedColor = new Color(0.8f, 0.4f, 0.4f);
	private final static Color worldColor = new Color(0.3f, 0.9f, 0.9f);
	private final static Color jointColor = new Color(0.5f, 0.8f, 0.8f);

	private float scale;
	private boolean isAntialiased;

	private HakoniwaEntityHolder holder;
	private HashSet<HakoniwaEntity> entities;
	private World world;
	private final Vec2[] worldBound = {
			new Vec2(),
			new Vec2(),
			new Vec2(),
			new Vec2()
	};

	private MouseJoint mouseJoint;
	private HakoniwaEntity mouseJointedEntity;
	final private Vec2 mouseJointPosition = new Vec2();

	// TODO implement viewport transform within this class.
	private OBBViewportTransform viewportTransform;
	private Graphics2D g;
	private BufferedImage image;
	private boolean updated = false;

	// Temporal variables used in public methods for converting coordinates.
	private final Vec2 realSrc = new Vec2();
	private final Vec2 screenDest = new Vec2();
	private final Vec2 realDest = new Vec2();
	private final Vec2 screenSrc = new Vec2();

	// Temporal variables used in methods for painting figures.
	private final Vec2 center = new Vec2();
	private final Vec2 tmp = new Vec2();
	private final Vec2 tmp1 = new Vec2();
	private final Vec2 tmp2 = new Vec2();
	private final Line2D.Float line = new Line2D.Float();
	private final GeneralPath path = new GeneralPath();
	final private Location location = new Location();
	final private Position position = new Position();
	final private ScreenLocation screenLocation = new ScreenLocation();
	final private ScreenPosition screenPosition = new ScreenPosition();

	public Hakoniwa() {

		// Instantiate the viewport.
		viewportTransform = new OBBViewportTransform();
		viewportTransform.setYFlip(true);

		// Set world bound.
		final AABB worldAABB = new AABB();
		final Vec2 worldLower = worldAABB.lowerBound;
		final Vec2 worldUpper = worldAABB.upperBound;
		worldLower.set(-WORLD_WIDTH/2, -WORLD_HEIGHT/2);
		worldUpper.set(WORLD_WIDTH/2, WORLD_HEIGHT/2);
		worldBound[0].set(worldLower.x, worldLower.y);
		worldBound[1].set(worldUpper.x, worldLower.y);
		worldBound[2].set(worldUpper.x, worldUpper.y);
		worldBound[3].set(worldLower.x, worldUpper.y);

		// Set gravity.
		final Vec2 gravity = new Vec2(0f, 0f);

		// Set whether to allow neglecting calculation of static objects.
		final boolean doSleep = true;

		// Instantiate the world.
		world = new World(worldAABB, gravity, doSleep);
		world.setAutoDebugDraw(false);
		world.setWarmStarting(true);
		world.setPositionCorrection(true);
		world.setContinuousPhysics(true);

		//
		setViewportSize(SCREEN_WIDTH, SCREEN_HEIGHT);
		setViewportScale(SCALE);
		entities = new HashSet<HakoniwaEntity>();
		holder = new HakoniwaEntityHolder(entities);
	}

	public void run() {
			for (HakoniwaEntity entity : entities) {
				entity.preStep();
				final Body body = entity.getBody();
				body.setLinearDamping(0.9f);
				body.setAngularDamping(0.9f);
			}
			synchronized (world) {
				world.step((float)getInterval()/1000, 20);
			}
			updated = true;
	}

	public void updateMouseJoint(Position p, HakoniwaEntity e) {
		synchronized (world) {
			if (mouseJoint != null && !mouseJointedEntity.equals(e)) {
				destroyMouseJoint();
			}
			mouseJointPosition.x = (float) p.getX()/100;
			mouseJointPosition.y = (float) p.getY()/100;
			if (mouseJoint == null) {
				final Body body = e.getBody();
				MouseJointDef md = new MouseJointDef();
				md.body1 = world.getGroundBody();
				md.body2 = body;
				md.target.set(mouseJointPosition);
				md.maxForce = 50f*body.m_mass;
				md.frequencyHz = 1000f/Matereal.getInstance().getInterval();
				md.dampingRatio = 0.9f;
				mouseJoint = (MouseJoint) world.createJoint(md);
				mouseJointedEntity = e;
			}
			mouseJoint.setTarget(mouseJointPosition);
		}
	}

	public void destroyMouseJoint() {
		synchronized (world) {
			if (mouseJoint != null) {
				world.destroyJoint(mouseJoint);
				mouseJoint = null;
				mouseJointedEntity = null;
			}
			return;
		}
	}

	public boolean hasMouseJoint() {
		return mouseJoint != null;
	}

	public HakoniwaEntity getMouseJointedEntity() {
		return mouseJointedEntity;
	}

	public World getWorld() {
		return world;
	}

	public HakoniwaEntityHolder getCorrespondingEntityHolder() {
		return holder;
	}

	public boolean isAntialiased() {
		return isAntialiased;
	}

	public void setAntialiased(boolean isAntialiased) {
		this.isAntialiased = isAntialiased;
		if (g != null) {
			g.setRenderingHint(
					RenderingHints.KEY_ANTIALIASING,
					isAntialiased ?
							RenderingHints.VALUE_ANTIALIAS_ON :
							RenderingHints.VALUE_ANTIALIAS_OFF);
		}
	}

	@Override
	public String getName() {
		return "Hakoniwa";
	}

	/**
	 * Set size of the viewport.
	 * @param screenWidth Width of the viewport in [px].
	 * @param screenHeight Height of the viewport in [px].
	 */
	public synchronized void setViewportSize(int screenWidth, int screenHeight) {
		viewportTransform.setExtents((float)screenWidth/2, (float)screenHeight/2);
		setViewportScale(scale);
		image = new BufferedImage(
				screenWidth,
				screenHeight,
				BufferedImage.TYPE_INT_ARGB);
		g = image.createGraphics();
		if (isAntialiased) {
			g.setRenderingHint(
					RenderingHints.KEY_ANTIALIASING,
					isAntialiased ?
							RenderingHints.VALUE_ANTIALIAS_ON :
							RenderingHints.VALUE_ANTIALIAS_OFF);
		}
	}

	/**
	 * Set scale of the viewport.
	 * @param scale [pixels/cm]
	 */
	public synchronized void setViewportScale(float scale) {
		this.scale = scale;
		viewportTransform.setCamera((float) getRealWidth()/200, (float) getRealHeight()/200, scale);
	}

	// CoordProvider

	public int getWidth() {
		return (int) viewportTransform.getExtents().x*2;
	}

	public int getHeight() {
		return (int) viewportTransform.getExtents().y*2;
	}

	public double getRealWidth() {
		return getWidth()/scale*100;
	}

	public double getRealHeight() {
		return getHeight()/scale*100;
	}

	public ScreenPosition realToScreen(Position realPosition) {
		final ScreenPosition screenPosition = new ScreenPosition();
		realToScreenOut(realPosition, screenPosition);
		return screenPosition;
	}

	public void realToScreenOut(Position realPosition, ScreenPosition screenPosition) {
		realSrc.x = (float) realPosition.getX()/100;
		realSrc.y = (float) realPosition.getY()/100;
		viewportTransform.getWorldToScreen(realSrc, screenDest);
		screenPosition.set((int) screenDest.x, (int) screenDest.y);
	}

	public Position screenToReal(ScreenPosition screenPosition)
			throws IllegalArgumentException {
		final Position realPosition = new Position();
		screenToRealOut(screenPosition, realPosition);
		return realPosition;
	}

	public void screenToRealOut(ScreenPosition screenPosition,
			Position realPosition) throws IllegalArgumentException {
		screenSrc.x = (float) screenPosition.getX();
		screenSrc.y = (float) screenPosition.getY();
		viewportTransform.getScreenToWorld(screenSrc, realDest);
		realPosition.set(realDest.x*100, realDest.y*100);
	}

	public void drawImage(Graphics g) {
		drawImage(g, 0, 0);
	}

	public void drawImage(Graphics g, int x, int y) {
		synchronized (image) {
			updateImage();
			g.drawImage(image, x, y, null);
		}
	}

	public BufferedImage getImage() {
		synchronized (image) {
			updateImage();
			final BufferedImage newImage = new BufferedImage(
					image.getWidth(), image.getHeight(), image.getType());
			final Graphics2D g2 = newImage.createGraphics();
			g2.drawImage(image, 0, 0, null);
			g2.dispose();
			return newImage;
		}
	}

	public byte[] getImageData() {
		synchronized (image) {
			updateImage();
			final BufferedImage newImage = new BufferedImage(
					image.getWidth(), image.getHeight(), BufferedImage.TYPE_3BYTE_BGR);
			final Graphics2D g2 = newImage.createGraphics();
			g2.setColor(Color.white);
			g2.fillRect(0, 0, getWidth(), getHeight());
			g2.drawImage(image, 0, 0, null);
			g2.dispose();
			return ((DataBufferByte)newImage.getRaster().getDataBuffer()).getData();
		}
	}

	private void updateImage() {
		if (!updated) {
			return;
		}
		synchronized (image){

			// Clear the image.
			g.setComposite(AlphaComposite.Clear);
			g.fillRect(0, 0, image.getWidth(), image.getHeight());
			g.setComposite(AlphaComposite.SrcOver);

			// Draw shapes.
			for (Body b = world.getBodyList(); b != null; b = b.getNext()) {
				final XForm xf = b.getMemberXForm();
				Color c = sleepingColor;
				if (b.m_userData instanceof HakoniwaEntity)
				{
					final HakoniwaEntity e = (HakoniwaEntity)b.m_userData;
					c = e.getColor();
				}
				for (Shape s = b.getShapeList(); s != null; s = s.getNext()) {
					drawShape(s, xf, c);
				}
			}

			// Draw joints.
			for (Joint j = world.getJointList(); j != null; j = j.getNext()) {
				drawJoint(j);
			}

			// Draw world bounds.
			setColor(worldColor);
			drawPolygon(worldBound, 4);
		}
		updated = false;
	}

	// EntityInformationProvider

	public Set<Entity> getEntities() {
		return new HashSet<Entity>(entities);
	}

	public boolean contains(Entity entity) {
		return entities.contains(entity);
	}

	// LocationProvider

	public CoordProvider getCoordProvider() {
		return this;
	}

	public Location getLocation(Entity e) {
		getLocationOut(e, location);
		return new Location(location);
	}

	public void getLocationOut(Entity e, Location location) {
		if (e instanceof HakoniwaEntity) {
			final Body body = ((HakoniwaEntity) e).getBody();
			final Vec2 position = body.getPosition();
			location.setLocation(
					(double) (position.x*100),
					(double) (position.y*100),
					(double) body.getAngle());
			return;
		}
		location.setLocation(0, 0, 0);
	}

	public Position getPosition(Entity e) {
		getLocationOut(e, location);
		return location.getPosition();
	}

	public void getPositionOut(Entity e, Position position) {
		getLocationOut(e, location);
		position.set(location.getX(), location.getY());
	}

	public double getX(Entity e) {
		getLocationOut(e, location);
		return location.getX();
	}

	public double getY(Entity e) {
		getLocationOut(e, location);
		return location.getY();
	}

	public double getRotation(Entity e) {
		getLocationOut(e, location);
		return location.getRotation();
	}

	// ScreenLocationProvider

	public ScreenLocation getScreenLocation(Entity e) {
		getScreenLocationOut(e, screenLocation);
		return new ScreenLocation(screenLocation);
	}

	public void getScreenLocationOut(Entity e, ScreenLocation screenLocation) {
		getLocationOut(e, location);
		position.set(location.getX(), location.getY());
		realToScreenOut(position, screenPosition);
		screenLocation.setLocation(
				screenPosition,
				-location.getRotation());
	}

	public ScreenPosition getScreenPosition(Entity e) {
		getScreenLocationOut(e, screenLocation);
		return screenLocation.getPosition();
	}

	public void getScreenPositionOut(Entity e, ScreenPosition screenPosition) {
		getScreenLocationOut(e, screenLocation);
		screenPosition.set(screenLocation.getX(), screenLocation.getY());
	}

	public int getScreenX(Entity e) {
		getScreenLocationOut(e, screenLocation);
		return screenLocation.getX();
	}

	public int getScreenY(Entity e) {
		getScreenLocationOut(e, screenLocation);
		return screenLocation.getY();
	}

	public double getScreenRotation(Entity e) {
		getScreenLocationOut(e, screenLocation);
		return screenLocation.getRotation();
	}

	// Painting features.

	private void setColor(Color color) {
		g.setColor(color);
	}

	private void drawShape(final Shape shape, final XForm xf, final Color color) {
		// This method uses center, tmp, tmp1 and tmp2 temporarily.
		setColor(color);
		switch (shape.getType())
		{

		case CIRCLE_SHAPE: {
			final CircleShape circle = (CircleShape) shape;
			XForm.mulToOut(xf, circle.getMemberLocalPosition(), center);
			viewportTransform.getWorldToScreen(center, tmp2);
			final float radius = circle.getRadius();
			final Vec2 axis = xf.R.col1;

			// Draw and fill a circle.
			float theta = 0.0f;
			path.reset();
			boolean isFirst = true;
			Vec2 first = tmp, current = tmp1;
			for (int i = 0; i < CIRCLE_SEGMENTS; ++i) {
				current.set(
						radius * MathUtils.cos(theta),
						radius * MathUtils.sin(theta));
				viewportTransform.vectorTransform(current, current);
				current.addLocal(tmp2);
				if (isFirst) {
					first.set(current);
					path.moveTo(current.x, current.y);
					isFirst = false;
				}
				else {
					path.lineTo(current.x, current.y);
				}
				theta += CIRCLE_THETA_INCREMENT;
			}
			path.lineTo(first.x, first.y);
			path.closePath();
			g.draw(path);
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.2f));
			g.fill(path);
			g.setComposite(AlphaComposite.SrcOver);

			// Draw an axis.
			tmp1.set(center.x + radius * axis.x, center.y + radius * axis.y);
			viewportTransform.getWorldToScreen(tmp1, tmp1);
			drawLineOnScreen(tmp1, tmp2);
			break;
		}

		case POINT_SHAPE: {
			final PointShape point = (PointShape) shape;
			XForm.mulToOut(xf, point.getMemberLocalPosition(), center);
			viewportTransform.getWorldToScreen(center, tmp2);

			// Draw a circle.
			float theta = 0.0f;
			Vec2 first = tmp, previous = null, current = tmp1;
			for (int i = 0; i < POINT_SEGMENTS; ++ i) {
				current.set(
						tmp2.x + MathUtils.cos(theta) * POINT_RADIUS,
						tmp2.y + MathUtils.sin(theta) * POINT_RADIUS);
				if (previous == null) {
					first.set(current);
					previous = tmp2;
				}
				else {
					drawLineOnScreen(previous, current);
				}
				previous.set(current);
				theta += POINT_THETA_INCREMENT;
			}
			drawLineOnScreen(current, first);
			break;
		}

		case POLYGON_SHAPE: {
			final PolygonShape poly = (PolygonShape) shape;
			final int vertexCount = poly.getVertexCount();
			final Vec2[] localVertices = poly.getVertices();
			if (vertexCount > Settings.maxPolygonVertices ||
					vertexCount <= 0) {
				break;
			}

			// Draw a polygon.
			Vec2 first = tmp, previous = null, current = tmp1;
			for (int i = 0; i < vertexCount; ++ i) {
				current = XForm.mul(xf, localVertices[i]);
				viewportTransform.getWorldToScreen(current, current);
				if (previous == null) {
					first = current;
				}
				else {
					drawLineOnScreen(previous, current);
				}
				previous = current;
			}
			drawLineOnScreen(current, first);
			break;
		}

		case EDGE_SHAPE:
			final EdgeShape edge = (EdgeShape) shape;
			XForm.mulToOut(xf, edge.getVertex1(), tmp1);
			XForm.mulToOut(xf, edge.getVertex2(), tmp2);
			drawLine(tmp1, tmp2);
			break;
		}
	}

	private void drawJoint(final Joint joint) {
		final Body b1 = joint.getBody1();
		final Body b2 = joint.getBody2();
		final XForm xf1 = b1.getMemberXForm();
		final XForm xf2 = b2.getMemberXForm();
		final Vec2 x1 = xf1.position;
		final Vec2 x2 = xf2.position;
		final Vec2 p1 = joint.getAnchor1();
		final Vec2 p2 = joint.getAnchor2();
		setColor(jointColor);
		switch (joint.getType())
		{

		case DISTANCE_JOINT:
			drawLine(p1, p2);
			break;

		case PULLEY_JOINT:
			final PulleyJoint pulley = (PulleyJoint) joint;
			final Vec2 s1 = pulley.getGroundAnchor1();
			final Vec2 s2 = pulley.getGroundAnchor2();
			drawLine(s1, p1);
			drawLine(s2, p2);
			drawLine(s1, s2);
			break;

		case MOUSE_JOINT:
			// Don't draw mouse joint
			break;

		case CONSTANT_VOLUME_JOINT:
			final ConstantVolumeJoint cvj = (ConstantVolumeJoint) joint;
			final Body[] bodies = cvj.getBodies();
			Vec2 first = null, previous = null, current = null;
			for (int i = 0; i < bodies.length; ++ i) {
				current = bodies[i].getMemberWorldCenter();
				viewportTransform.getWorldToScreen(current, current);
				if (previous == null) {
					first = current;
				}
				else {
					drawLineOnScreen(previous, current);
				}
				previous = current;
			}
			drawLineOnScreen(current, first);
			break;

		default:
			drawLine(x1, p1);
			drawLine(p1, p2);
			drawLine(x2, p2);
			break;
		}
	}

	private void drawPolygon(Vec2[] vertices, int vertexCount) {
		Vec2 first = null, previous = null, current = null;
		for (int i = 0; i < vertexCount; ++ i) {
			current = vertices[i];
			if (previous == null) {
				first = current;
			}
			else {
				drawLine(previous, current);
			}
			previous = current;
		}
		drawLine(current, first);
	}

	private void drawLine(Vec2 argP1, Vec2 argP2) {
		viewportTransform.getWorldToScreen(argP1, tmp1);
		viewportTransform.getWorldToScreen(argP2, tmp2);
		drawLineOnScreen(tmp1, tmp2);
	}

	private void drawLineOnScreen(Vec2 p1, Vec2 p2) {
		line.setLine(p1.x, p1.y, p2.x, p2.y);
		g.draw(line);
	}

}
